@@ -18,7 +18,7 @@
 
 /* Written by Marc Tardif <intmktg@cam.org>.  */
 
-/* The buffering algorithm in this program can be summarised with the
+/* The buffering algorithm in this program can be summarized with the
    following illustration:
 
    buffer       pos_beg    buf_beg         pos_end   buf_end     lim
@@ -53,16 +53,16 @@
    Manipulation of the input consists of maintaining two pointers:
    - The buf_end address points to the last character parsed.  When
      the current buffer is exhausted and buf_end is not equal to lim,
-     the difference consitutes the residue for the next input.
+     the difference constitutes the residue for the next input.
    - When parsing text, the pos_end address is always the same as
      buf_end.  But when parsing HTML, the formatted text is copied
      over the current buffer and the pointer is stored in pos_end.
      When lim is reached, the save buffer is copied from this point
      backwards.
 
-   Note that pos_end and pos_beg are actually the same variable in
-   the source code, one becomes the other after filling the input
-   buffer.  Same applies to buf_end and buf_beg.
+   Note that pos_end and pos_beg are actually the same variable in the
+   source code, one becomes the other after filling the input buffer.
+   Same applies to buf_end and buf_beg.
 */
 
 #include <sys/types.h>
@@ -80,21 +80,16 @@
 extern int errno;
 #endif
 
-#ifndef HAVE_STRERROR
-extern int sys_nerr;
-extern char *sys_errlist[];
-# define strerror(E) (0 <= (E) && (E) < sys_nerr ? sys_errlist[E] : "Unknown system error")
-#endif
-
 #if ! defined HAVE_MEMMOVE && ! defined memmove
 # define memmove(d, s, n) bcopy (s, d, n)
 #endif
 
-#include "kw.h"
+#include "context.h"
+#include "err.h"
 #include "getopt.h"
 #include "options.h"
 
-#define BUFFER 8192
+#define BUFFER 4096
 #define SAVE    512
 
 /* Base of buffer.  */
@@ -105,9 +100,10 @@ static size_t bufsalloc;
 
 /* Command-line options.  */
 struct options opt;
-static char const short_options[] = "C:D:FHKLO:Vchiklq";
+static char const short_options[] = "C:D:FHLO:PVbchilnpqs";
 static struct option long_options[] =
 {
+  {"byte-offset", no_argument, NULL, 'b'},
   {"count", no_argument, NULL, 'c'},
   {"context", required_argument, NULL, 'C'},
   {"distance", required_argument, NULL, 'D'},
@@ -116,34 +112,19 @@ static struct option long_options[] =
   {"fixed-string", no_argument, NULL, 'F'},
   {"help", no_argument, (int *)&opt.show_help, 1},
   {"ignore-case", no_argument, NULL, 'i'},
+  {"line-number", no_argument, NULL, 'n'},
   {"no-filename", no_argument, NULL, 'h'},
-  {"no-keyword", no_argument, NULL, 'k'},
-  {"occurences", required_argument, NULL, 'O'},
+  {"no-pattern", no_argument, NULL, 'p'},
+  {"occurrences", required_argument, NULL, 'O'},
   {"quiet", no_argument, NULL, 'q'},
   {"silent", no_argument, NULL, 'q'},
+  {"no-messages", no_argument, NULL, 's'},
   {"version", no_argument, NULL, 'V'},
   {"with-filename", no_argument, NULL, 'H'},
-  {"with-keyword", no_argument, NULL, 'K'},
+  {"with-pattern", no_argument, NULL, 'P'},
   {0, 0, 0, 0}
 };
 
-static void
-syserr (const char *msg)
-{
-  if (errno)
-    fprintf (stderr, "bool: %s: %s\n", msg, strerror (errno));
-  else
-    fprintf (stderr, "bool: %s\n", msg);
-  exit (1);
-}
-
-static void
-fatal (const char *msg)
-{
-  fprintf (stderr, "bool: %s\n", msg);
-  exit (2);
-}
-
 static int
 ck_atoi (char const *str, int *out)
 {
@@ -157,29 +138,26 @@ ck_atoi (char const *str, int *out)
   return 0;
 }
 
-static int
+static void
 init (char *str)
 {
   buffer = malloc (BUFFER + SAVE + 1);
   if (!buffer)
-    return 0;
+    err_fatal ("memory exhausted");
 
-  switch (kw_init (str))
+  switch (context_init (str))
     {
-    case -1:
-      fatal ("invalid query");
     case 0:
-      return 0;
+      err_fatal ("invalid query");
     case 1:
       break;
     default:
-      if (!opt.no_keyword)
-        opt.with_keyword = 1;
+      if (!opt.no_pattern)
+        opt.with_pattern = 1;
       break;
     }
 
   bufsalloc = SAVE;
-  return 1;
 }
 
 static enum extension
@@ -188,9 +166,13 @@ get_extension (char *file)
   register char *end;
   struct { char *str; int val; } ext[] =
     {
-      { "html", HTML },
-      { "htm", HTML },
-      { NULL, TEXT }
+      { "c",    SOURCE },
+      { "cpp",  SOURCE },
+      { "h",    SOURCE },
+      { "pl",   SOURCE },
+      { "html", HTML   },
+      { "htm",  HTML   },
+      { NULL,   TEXT   }
     };
 
   file += 1;
@@ -218,7 +200,7 @@ get_extension (char *file)
 }
 
 static int
-readfile (int fd)
+file_read (int fd)
 {
   ssize_t bytesread;
   size_t residue, save;
@@ -241,7 +223,7 @@ readfile (int fd)
 
       lim = buf + bytesread;
       *lim = '\0';
-      save = kw_exec (&pos, &buf, lim);
+      save = context_find (&pos, &buf, lim);
       residue = lim - buf;
 
       residue_offset = buf - buffer;
@@ -253,7 +235,7 @@ readfile (int fd)
 
           buffer = realloc (buffer, bufsalloc + BUFFER + 1);
           if (!buffer)
-            fatal ("memory exhausted");
+            err_fatal ("memory exhausted");
         }
 
       buf = buffer + bufsalloc;
@@ -269,14 +251,14 @@ readfile (int fd)
   while (!bytesread && pos < lim)
     {
       buf = lim - 1;
-      kw_exec (&pos, &buf, lim);
+      context_find (&pos, &buf, lim);
     }
 
-  return kw_print ();
+  return context_print ();
 }
 
 static int
-openfile (char *file)
+file_open (char *file)
 {
   enum extension type;
   int desc, status;
@@ -292,17 +274,23 @@ openfile (char *file)
       while ((desc = open (file, O_RDONLY)) < 0 && errno == EINTR)
         continue;
 
+      if (desc < 0)
+        {
+          err_message ("open");
+          return 0;
+        }
+
       type = get_extension (file);
       opt.filename = file;
     }
 
-  kw_prep (type);
-  status = readfile (desc);
+  context_prep (type);
+  status = file_read (desc);
 
   while (close (desc) != 0)
     if (errno != EINTR)
       {
-        syserr ("close");
+        err_message ("close");
         break;
       }
 
@@ -323,14 +311,16 @@ Usage: bool [OPTION]... PATTERN [FILE] ...\n\
 Search for PATTERN in each FILE or standard input.\n\
 \n\
 Interpretation:\n\
-  -F, --fixed-string        PATTERN is a string, not a boolean expression\n\
+  -F, --fixed-string        PATTERN is a string, not an expression\n\
   -i, --ignore-case         ignore case distinctions\n\
 \n\
 Output control:\n\
+  -b, --byte-offset         print the byte offset with output lines\n\
+  -n, --line-number         print line number with output lines\n\
   -H, --with-filename       print the filename for each match\n\
   -h, --no-filename         suppress the prefixing filename on output\n\
-  -K, --with-keyword        print the keyword for each match\n\
-  -k, --no-keyword          suppress the prefixing keyword on outptut\n\
+  -P, --with-pattern        print the pattern for each match\n\
+  -p, --no-pattern          suppress the prefixing pattern on outptut\n\
   -q, --quiet, --silent     suppress all normal output\n\
   -L, --files-without-match only print FILE names containing no match\n\
   -l, --files-with-matches  only print FILE names containing matches\n\
@@ -339,9 +329,10 @@ Output control:\n\
 Context control:\n\
   -C, --context=NUM         print NUM characters of output context\n\
   -D, --distance=NUM        NEAR matches are at most NUM words apart\n\
-  -O, --occurences=NUM      print NUM lines of context for each keyword\n\
+  -O, --occurrences=NUM     print NUM lines of context for each pattern\n\
 \n\
 Miscellaneous:\n\
+  -s, --no-messages         suppress error messages\n\
   -V, --version             print version information and exit\n\
       --help                display this help and exit\n\
 ");
@@ -359,7 +350,7 @@ main (int argc, char *argv[])
   state = 0;
   opt.context = DEFAULT_CONTEXT;
   opt.distance = DEFAULT_DISTANCE;
-  opt.occurences = DEFAULT_OCCURENCES;
+  opt.occurrences = DEFAULT_OCCURRENCES;
 
   setlocale (LC_CTYPE, "");
 
@@ -367,11 +358,14 @@ main (int argc, char *argv[])
          != -1)
     switch (c)
       {
+      case 'b':
+        opt.out_byte = 1;
+        break;
       case 'C':
         if (optarg)
           {
             if (ck_atoi (optarg, &opt.context))
-              fatal ("invalid context argument");
+              err_fatal ("invalid context argument");
           }
         break;
       case 'c':
@@ -382,7 +376,7 @@ main (int argc, char *argv[])
         if (optarg)
           {
             if (ck_atoi (optarg, &opt.distance))
-              fatal ("invalid distance argument");
+              err_fatal ("invalid distance argument");
           }
         break;
       case 'F':
@@ -397,12 +391,6 @@ main (int argc, char *argv[])
       case 'i':
         opt.ignore_case = 1;
         break;
-      case 'K':
-        opt.with_keyword = 1;
-        break;
-      case 'k':
-        opt.no_keyword = 1;
-        break;
       case 'L':
         opt.out_quiet = 1;
         opt.list_files = 1;
@@ -411,16 +399,28 @@ main (int argc, char *argv[])
         opt.out_quiet = 1;
         opt.list_files = 2;
         break;
+      case 'n':
+        opt.out_line = 1;
+        break;
       case 'O':
         if (optarg)
           {
-            if (ck_atoi (optarg, &opt.occurences))
-              fatal ("invalid occurence argument");
+            if (ck_atoi (optarg, &opt.occurrences))
+              err_fatal ("invalid occurrence argument");
           }
         break;
+      case 'P':
+        opt.with_pattern = 1;
+        break;
+      case 'p':
+        opt.no_pattern = 1;
+        break;
       case 'q':
         opt.out_quiet = 1;
         break;
+      case 's':
+        opt.suppress_errors = 1;
+        break;
       case 'V':
         printf ("bool (GNU) %s\n", VERSION);
         exit (0);
@@ -438,8 +438,7 @@ main (int argc, char *argv[])
   if (optind == argc)
     usage (1);
 
-  if (!init (argv[optind++]))
-    fatal ("memory exhausted");
+  init (argv[optind++]);
 
   if (argc - optind > 1 && !opt.no_filename)
     opt.with_filename = 1;
@@ -450,14 +449,16 @@ main (int argc, char *argv[])
       do
         {
           char *file = argv[optind];
-          state += openfile (strcmp (file, "-") == 0 ? (char *) NULL : file);
+          state += file_open (strcmp (file, "-") == 0
+            ? (char *) NULL
+            : file);
         }
       while (++optind < argc);
     }
   else
-    state = openfile ((char *) NULL);
+    state = file_open ((char *) NULL);
 
-  kw_free ();
+  context_free ();
   free (buffer);
   return !state;
 }
