@@ -18,7 +18,7 @@
 
 /* Written by Marc Tardif <intmktg@cam.org>.  */
 
-/* The memory allocation algorithm in this program can be summarised
+/* The memory allocation algorithm in this program can be summarized
    with the following illustration:
 
     contents      next                                           lim
@@ -27,10 +27,10 @@
       *
    pointer
 
-   The pointer variable contains the address of the previously allo-
-   cated object.  The first pointer is NULL to identify the object
-   as the last element in the list.  This creates a singly linked
-   list which can be traversed backwards.
+   The pointer variable contains the address of the previously allocated
+   object.  The first pointer is NULL to identify the object as the last
+   element in the list.  This creates a singly linked list which can be
+   traversed backwards.
 
    The memory block starts immediately after the pointer variable at
    the address of contents.  For the current active block of memory
@@ -38,18 +38,17 @@
    - next, the address of the next available character.
    - lim, the end of the current memory block.
 
-   Both the pointer and memory block are allocated in a single call
-   to malloc.  This is accomplished by requesting memory past the
-   contents variable and manipulating this area with the pointers
-   mentioned above.
+   Both the pointer and memory block are allocated in a single call to
+   malloc.  This is accomplished by requesting memory past the contents
+   variable and manipulating this area with the pointers mentioned above.
 */
 
 #include <stdlib.h>
+#include <unistd.h>
 
+#include "err.h"
 #include "mem.h"
 
-#define BLOCK_SIZE 4096
-
 /* Block of memory distributed in segments.                              */
 struct block
 {
@@ -63,25 +62,28 @@ struct mem
   void *block;                /* Address of current struct block.        */
   char *lim;                  /* Address of char after current block.    */
   char *next;                 /* Next char to current block.             */
+  size_t size;                /* Size of memory block.                   */
 };
 
 static struct mem *mem;
 
-void *
-mem_init (int extra)
+void
+mem_init (void)
 {
   register struct block *block;
+  size_t page_size;
 
-  mem = malloc (ALIGN (sizeof (struct mem)) + BLOCK_SIZE + extra);
+  page_size = getpagesize ();
+  mem = malloc (page_size);
   if (!mem)
-    return 0;
+    err_fatal ("memory exhausted");
 
   block = mem->block = (struct block *)
-                         ((char *)mem + ALIGN (sizeof (struct mem)));
-  mem->next = block->contents + extra;
-  mem->lim = (char *) block + BLOCK_SIZE + extra;
+      ((char *)mem + ALIGN (sizeof (struct mem)));
+  mem->next = block->contents;
+  mem->lim = (char *) mem + page_size;
+  mem->size = page_size;
   block->prev = NULL;
-  return mem->next - extra;
 }
 
 void *
@@ -94,12 +96,13 @@ mem_alloc (int length)
       register struct block *block;
       register long new_size;
 
+      mem->size *= 2;
       new_size = length + sizeof (struct block *);
-      new_size += BLOCK_SIZE - (new_size % BLOCK_SIZE);
+      new_size += mem->size - (new_size % mem->size);
 
       block = malloc (new_size);
       if (!block)
-        return NULL;
+        err_fatal ("memory exhausted");
 
       block->prev = mem->block;
       mem->block = block;
