@@ -24,7 +24,7 @@
 #define ALIGNBYTES      (sizeof (int) - 1)
 #define ALIGN(p)        (((unsigned)(p) + ALIGNBYTES) & ~ALIGNBYTES)
 
-extern void *mem_init  (int);
+extern void  mem_init  (void);
 extern void *mem_alloc (int);
 extern void  mem_free  (void);
 
