@@ -18,13 +18,14 @@
 
 /* Written by Marc Tardif <intmktg@cam.org>.  */
 
-/* The routines in this program support the SGML constructs that are
+/*
+   The routines in this program support the SGML constructs that are
    used in the HTML 4.01 standard (http://www.w3c.org/TR/html4/).
 
-   An SGML document type definition declares element types that
-   represent structures and desired behavior.  For pattern matching,
-   only the element types affecting the spacing between display
-   characters are considered.  Some common examples include:
+   An SGML document type definition declares element types that represent
+   structures and desired behavior.  For pattern matching, only the
+   element types affecting the spacing between display characters are
+   considered.  Some common examples include:
    - "<p>" represents a new paragraph.
    - "<br>" represents a new line.
    - "<sup>" represents a space.
@@ -185,7 +186,8 @@ get_entity (char **str, char *lim)
     return '&';
 
   *str = s;
-  return entitylist[((hash & 0x3F) ^ entityhash[hash >> 26]) % ENTITY_NUM].val;
+  return entitylist[((hash & 0x3F) ^ entityhash[hash >> 26])
+    % ENTITY_NUM].val;
 }
 
 unsigned char
