@@ -18,47 +18,48 @@
 
 /* Written by Marc Tardif <intmktg@cam.org>.  */
 
-/* The state machine in this program can be summarised with the
-   following components:
+/* The state machine in this program can be summarized with the following
+   components:
 
-   - A list of events that are handled by the machine.  These are
-     actually the characters read from the input which can take
-     values from 0 to UCHAR_MAX.  The NUL character is special
-     because it is also used as a sentinel to mark the buffer end.
+   - A list of events that are handled by the machine.  These are actually
+     the characters read from the input which can take values from 0 to
+     UCHAR_MAX.  The NUL character is special because it is also used as
+     a sentinel to mark the buffer end.
 
-   - A list of actions that describe how the machine reacts to an
-     event.  The return value is used in a multi-way decision tree
-     to determine the statements to run on a particular action.
+   - A list of actions that describe how the machine reacts to an event.
+     The return value is used in a multi-way decision tree to determine
+     the statements to run on a particular action.
 
-   - A set of states for each possible condition of the machine.
-     Each state corresponds to a table containing as many elements
-     as events.  For each event in a table, a corresponding action
-     is defined.
+   - A set of states for each possible condition of the machine.  Each
+     state corresponds to a table containing as many elements as events.
+     For each event in a table, a corresponding action is defined.
 
    The list of events can be separated in subsets which constantly
    share the same actions.  For example, numeric characters always
-   return the same action for a particular state.  Therefore, a
-   GENERIC table is defined for the set of possible events and
-   returns a subset.  This smaller set can then be used to define
-   smaller state tables.
+   return the same action for a particular state.  Therefore, a GENERIC
+   table is defined for the set of possible events and returns a subset.
+   This smaller set can then be used to define smaller state tables.
 */
 
 #include <stdlib.h>
 
-#include "ac.h"
-#include "kw.h"
+#include "context.h"
+#include "match.h"
 #include "options.h"
 
-#include "tab.h"       /* GENERIC state table.  */
-#include "text.h"      /* TEXT state table.  */
+#include "event.h"
+#include "text.h"
 
 enum
 {
   NOP,  /* no operation                         */
   PAT,  /* pattern                              */
+  L2F,  /* line to form feed                    */
+  L2L,  /* line to line                         */
+  PL2F, /* pattern-line to form feed            */
+  PL2L, /* pattern-line to line                 */
   PL2P, /* pattern-line to pattern              */
   PL2D, /* pattern-line to dash                 */
-  PL2S, /* pattern-line to space                */
   PS2P, /* pattern-space to pattern             */
   PS2D, /* pattern-space to dash                */
   PS2F, /* pattern-space to form feed           */
@@ -67,9 +68,12 @@ enum
   PW2F, /* pattern to form feed                 */
   PW2L, /* pattern to line                      */
   PW2S, /* pattern to space                     */
+  S2F,  /* space to form feed                   */
+  S2L,  /* space to line                        */
   S2N,  /* space to NUL byte                    */
   S2P,  /* space to pattern                     */
   S2W,  /* space to word                        */
+  W2L,  /* word to line                         */
   W2N,  /* word to NUL byte                     */
   W2P,  /* word to pattern                      */
   W2S,  /* word to space                        */
@@ -84,7 +88,7 @@ static unsigned char pspace[] =
           PS2P, PS2P, PS2P, PS2P, PS2D, PS2D, PS2P, PS2P,
           PS2P, PS2P, PS2P, PS2P, PS2P, PS2P, PS2P, PS2P };
 static unsigned char space[] =
-  {  S2N,  NOP,  NOP,  NOP,  NOP,  NOP,  NOP,  S2W,  S2P,
+  {  S2N,  NOP,  NOP,  S2L,  S2L,  S2F,  S2F,  S2W,  S2P,
            S2W,  S2P,  S2W,  S2P,  S2W,  S2P,  S2W,  S2P,
            S2W,  S2P,  S2W,  S2P,  S2W,  S2P,  S2W,  S2P };
 static unsigned char pword[] =
@@ -92,15 +96,19 @@ static unsigned char pword[] =
            PAT,  PAT,  PAT,  PAT, PW2D, PW2D,  PAT,  PAT,
            PAT,  PAT,  PAT,  PAT,  PAT,  PAT,  PAT,  PAT };
 static unsigned char word[] =
-  {  W2N,  W2S,  W2S,  W2S,  W2S,  W2S,  W2S,  WRD,  W2P,
+  {  W2N,  W2S,  W2S,  W2L,  W2L,  W2S,  W2S,  WRD,  W2P,
            WRD,  W2P,  WRD,  W2P,  WRD,  W2P,  WRD,  W2P,
            WRD,  W2P,  WRD,  W2P,  WRD,  W2P,  WRD,  W2P };
 static unsigned char pline[] =
-  {  S2N,  NOP,  NOP, PL2S, PL2S, PL2S, PL2S, PL2P, PL2P,
+  {  S2N,  NOP,  NOP, PL2L, PL2L, PL2F, PL2F, PL2P, PL2P,
           PL2P, PL2P, PL2P, PL2P, PL2D, PL2D, PL2P, PL2P,
           PL2P, PL2P, PL2P, PL2P, PL2P, PL2P, PL2P, PL2P };
+static unsigned char line[] =
+  {  S2N,  NOP,  NOP,  L2L,  L2L,  L2F,  L2F,  S2W,  S2P,
+           S2W,  S2P,  S2W,  S2P,  S2W,  S2P,  S2W,  S2P,
+           S2W,  S2P,  S2W,  S2P,  S2W,  S2P,  S2W,  S2P };
 
-unsigned char *text_first = word;
+unsigned char *text_first = space;
 
 int
 text_find (char **pos, char **buf, char *lim)
@@ -108,16 +116,18 @@ text_find (char **pos, char **buf, char *lim)
   register unsigned char c;
   register char *end;
   unsigned char *table;
-  int ret, words;
+  int ret, num_paragraphs, num_lines, num_words;
 
-  table = *kw->pattern_table;
-  words = kw->words;
+  table = *context->pattern_table;
+  num_paragraphs = context->num_paragraphs;
+  num_lines = context->num_lines;
+  num_words = context->num_words;
   end = *pos;
 
   for (;;)
     {
       c = *end++;
-      switch (table[tab[c]])
+      switch (table[event[c]])
         {
         case S2N:
         case W2N:
@@ -131,20 +141,36 @@ text_find (char **pos, char **buf, char *lim)
         case NOP:
           break;
 
-        case PL2S:
+        case PL2L:
+          num_paragraphs++;
+        case S2L:
+        case W2L:
+          num_lines++;
+          table = line;
+          break;
+
+        case L2L:
+          num_lines++;
+        case L2F:
+        case S2F:
+          num_paragraphs++;
+          break;
+
+        case PL2F:
         case PS2F:
         case PW2F:
+          num_paragraphs++;
         case W2S:
           table = space;
           break;
 
         case S2W:
-          words++;
+          num_words++;
           table = word;
           break;
 
         case S2P:
-          words++;
+          num_words++;
         case W2P:
           table = pword;
           find_first ();
@@ -152,7 +178,7 @@ text_find (char **pos, char **buf, char *lim)
 
         case PS2D: case PL2D:
         case PS2P: case PL2P:
-          words++;
+          num_words++;
           table = pword;
         case PAT:
           find_next ();
@@ -163,6 +189,7 @@ text_find (char **pos, char **buf, char *lim)
           find_space ();
           break;
         case PW2L:
+          num_lines++;
           table = pline;
           find_space ();
           break;
@@ -173,6 +200,7 @@ text_find (char **pos, char **buf, char *lim)
           break;
 
         case PS2L:
+          num_lines++;
           table = pline;
           break;
 
@@ -182,8 +210,11 @@ text_find (char **pos, char **buf, char *lim)
     }
 
  fin:
-  kw->words = words;
-  *kw->pattern_table = table;
+  context->num_bytes += end - *pos;
+  context->num_lines = num_lines;
+  context->num_words = num_words;
+  context->num_paragraphs = num_paragraphs;
+  *context->pattern_table = table;
   *buf = *pos = end;
 
   return ret - 1;
@@ -197,15 +228,15 @@ text_forward (char *pos, char *lim)
   unsigned char *table;
   char *max;
 
-  table = *kw->context_table ? *kw->context_table : pword;
-  max = kw->str + MAX (opt.context/2, kw->bpos - 1);
-  fpos = kw->str + kw->fpos;
+  table = *context->forward_table ? *context->forward_table : pword;
+  max = context->tmp->str + MAX (opt.context / 2, context->tmp->bpos - 1);
+  fpos = context->tmp->str + context->tmp->fpos;
   end = pos;
 
   while (fpos < max)
     {
       c = *end++;
-      switch (table[tab[c]])
+      switch (table[event[c]])
         {
         case S2N:
           if (end <= lim)
@@ -238,7 +269,8 @@ text_forward (char *pos, char *lim)
           table = pline;
           break;
 
-        case PL2S:
+        case PL2L:
+        case PL2F:
         case PS2F:
           fpos--;
         case PW2F:
@@ -255,8 +287,8 @@ text_forward (char *pos, char *lim)
   table = NULL;
 
  fin:
-  *kw->context_table = table;
-  kw->fpos = fpos - kw->str;
+  *context->forward_table = table;
+  context->tmp->fpos = fpos - context->tmp->str;
   return table;
 }
 
@@ -267,14 +299,14 @@ text_backward (char *pos)
   register char *beg, *bpos;
   unsigned char *table;
 
-  bpos = kw->str + opt.context;
+  bpos = context->tmp->str + opt.context;
   table = pspace;
   beg = pos;
 
-  while (bpos > kw->str)
+  while (bpos > context->tmp->str)
     {
       c = *(--beg);
-      switch (table[tab[c]])
+      switch (table[event[c]])
         {
         case PAT:
         case PW2D:
@@ -304,7 +336,8 @@ text_backward (char *pos)
           pos = beg + 1;
           goto fin;
 
-        case PL2S:
+        case PL2L:
+        case PL2F:
         case PS2F:
         case S2N:
           bpos++;
@@ -317,7 +350,7 @@ text_backward (char *pos)
     }
 
  fin:
-  kw->bpos = bpos - kw->str + 1;
+  context->tmp->bpos = bpos - context->tmp->str + 1;
   return pos;
 }
 
